#define _GNU_SOURCE

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <execinfo.h>
#include <stdint.h>
#include <signal.h>
#include <time.h>
#include <tiffio.h>


Pixmap (*RealXCreatePixmap)(Display *display, 
			    Drawable d, 
			    unsigned int width, 
			    unsigned int height,
			    unsigned int depth);

int (*RealXFreePixmap)(Display *display,
		       Pixmap p);

Display* (*RealXOpenDisplay)(const char *name);
int (*RealXCloseDisplay)(Display *d);

char *logdir;
FILE *logfile;

typedef struct _PixmapInfo {
  Pixmap p;
  int width;
  int height;
  int depth;
  time_t alloced_at;
  time_t freed_at;
} PixmapInfo;

void savePixmap(Display *,
		int,
		PixmapInfo *);

void flushPixmaps(Display *d,
		  int id,
		  int force);
void doSignalUSR1(int);
void doSignalUSR2(int);
void doExit(void);

static int save_at_checkpoint = 0;
static int ndisplays = 0;
static Display **displays = NULL;
static int *npixmaps = NULL;
static PixmapInfo **pixmaps = NULL;


void do_startup(void) __attribute__ ((constructor));
void do_startup(void)  {
  char *policy;
  if ((RealXOpenDisplay = (Display* (*)(const char *))dlsym(RTLD_NEXT, "XOpenDisplay")) == NULL) {
    fprintf(stderr, "Cannot resolve real XOpenDisplay function");
    exit(1);
  }

  if ((RealXCloseDisplay = (int (*)(Display *))dlsym(RTLD_NEXT, "XCloseDisplay")) == NULL) {
    fprintf(stderr, "Cannot resolve real XCloseDisplay function");
    exit(1);
  }

  if ((RealXCreatePixmap = (Pixmap (*)(Display*, Drawable, unsigned int, unsigned int, unsigned int))dlsym(RTLD_NEXT, "XCreatePixmap")) == NULL) {
    fprintf(stderr, "Cannot resolve real XCreatePixmap function");
    exit(1);
  }
  
  if ((RealXFreePixmap = (int (*)(Display*, Pixmap))dlsym(RTLD_NEXT, "XFreePixmap")) == NULL) {
    fprintf(stderr, "Cannot resolve real XFreePixmap function");
    exit(1);
  }

  if ((policy = getenv("CAPTURE_PIXMAP_SAVE_POLICY")) != NULL) {
    if (!strncmp(policy, "checkpoint", 10)) {
      save_at_checkpoint = 1;
    }
  }
}


void trace(FILE *log) {
  void *trace[100];
  int i; 
  int depth = backtrace(trace, 100);
  
  for (i = 0 ; i < depth ; i++) {
    Dl_info dlinfo;
    if (!(dladdr(trace[i], &dlinfo))) {
      fprintf(stderr, "cannot resolve library containing %p\n", trace[i]);
    }
    
    fprintf(log, "| %p %p:%s (%p+%p:%s)\n", trace[i], dlinfo.dli_saddr, dlinfo.dli_sname, dlinfo.dli_fbase, (void*)(trace[i] - dlinfo.dli_fbase), dlinfo.dli_fname); 
  }
}


Display *XOpenDisplay(const char *name) {
  Display *d = RealXOpenDisplay(name);
  
  if (d != NULL) {
    if (ndisplays == 0) {
      char dir[1024];
      char file[1024];
      if (getenv("CAPTURE_DIR")) {
	logdir = getenv("CAPTURE_DIR");
      } else {
	logdir = "capture-pixmap";
      }
      snprintf(dir, 1024, "%s-%d", logdir, getpid());
      mkdir(dir, 0755);
      snprintf(file, 1024, "%s/pixmaps.log", dir);

      if (!(logfile = fopen(file, "w"))) {
	fprintf(stderr, "Cannot open logfile %s for writing\n", file);
	exit(1);
      }

      signal(SIGUSR1, doSignalUSR1);
      signal(SIGUSR2, doSignalUSR2);
      atexit(doExit);
    }

    ndisplays++;
    displays = realloc(displays, sizeof(Display*)*ndisplays);
    displays[ndisplays-1] = d;
    npixmaps = realloc(npixmaps, sizeof(int)*ndisplays);
    npixmaps[ndisplays-1] = 0;
    pixmaps = realloc(pixmaps, sizeof(PixmapInfo*)*ndisplays);
    pixmaps[ndisplays-1] = NULL;
  }
  
  return d;
}

int lookupDisplay(Display *d) {
  int i;
  for (i = 0 ; i < ndisplays ; i++) {
    if (displays[i] == d) {
      return i;
    }
  }
  return -1;
}

int XCloseDisplay(Display *d) {
  int id = lookupDisplay(d);
  
  if (id != -1) {
    int i;
    for (i = 0; i < ndisplays ; i++) {
      if (i > id) {
	displays[i-1] = displays[i];
	npixmaps[i-1] = npixmaps[i];
	free(pixmaps[i]);
	pixmaps[i-1] = pixmaps[i];
      }
    }
  }

  if (id != -1) {
    if (!save_at_checkpoint) {
      flushPixmaps(displays[id], id, 1);
    }
    ndisplays--;
    
    if (ndisplays > 0) {
      displays = realloc(displays, sizeof(Display *)*ndisplays);
      npixmaps = realloc(npixmaps, sizeof(int)*ndisplays);
      pixmaps = realloc(pixmaps, sizeof(PixmapInfo*)*ndisplays);
    } else {
      free(displays);
      free(npixmaps);
      free(pixmaps);
      displays = NULL;
      npixmaps = NULL;
      pixmaps = NULL;
    }
  }
  
  return RealXCloseDisplay(d);
}

Pixmap XCreatePixmap(Display *display, 
		     Drawable d, 
		     unsigned int width, 
		     unsigned int height,
		     unsigned int depth) {
  int id = lookupDisplay(display);
  Pixmap p = RealXCreatePixmap(display, d, width, height, depth);
  time_t now = time(NULL);

  npixmaps[id]++;

  pixmaps[id] = realloc(pixmaps[id], sizeof(PixmapInfo)*npixmaps[id]);
  
  pixmaps[id][npixmaps[id]-1].p = p;
  pixmaps[id][npixmaps[id]-1].width = width;
  pixmaps[id][npixmaps[id]-1].height = height;
  pixmaps[id][npixmaps[id]-1].depth = depth;
  if ((pixmaps[id][npixmaps[id]-1].alloced_at = now) < 0) {
    perror("time");
    exit(1);
  }
  pixmaps[id][npixmaps[id]-1].freed_at = 0;

  fprintf(logfile, "+ %lu %dx%d@%d t=%ld\n", p, width, height, depth, now);
  trace(logfile);
  fflush(logfile);

  return p;
}

int XFreePixmap(Display *display,
		Pixmap p) {
  int id = lookupDisplay(display);
  time_t now = time(NULL);
  int i;

  fprintf(logfile, "- %lu t=%ld\n", p, now);
  trace(logfile);
  fflush(logfile);

  for (i = 0 ; i < npixmaps[id] ; i++) {
    if (pixmaps[id][i].p == p) {
      if (!save_at_checkpoint) {
	savePixmap(display, id, &pixmaps[id][i]);
      }
  
      if ((pixmaps[id][i].freed_at = now) < 0) {
	perror("time");
	exit(1);
      }

      break;
    }
  }
  
  return RealXFreePixmap(display, p);
}

unsigned int ones(unsigned long bitmask)
 {
  unsigned int i=0;
  
  while (bitmask!=0)
   {
    if ((bitmask & 1) == 1)
     i++;
    bitmask=bitmask/2;
   };

  return i;
 };

void savePixmap(Display *display,
		int id,
		PixmapInfo *p) {  
  XVisualInfo tmplVisualInfo;
  XVisualInfo *myVisualInfo;
  int visuals;
  int redShift = 0, greenShift = 0, blueShift = 0;
  XImage *img;
  char filename[1024];
  TIFF* out;
  int row;
  struct stat st;
  //int sampleperpixel = (img->depth / 8);
  int sampleperpixel = 3;
  //tsize_t linebytes  = img->bytes_per_line;
  tsize_t linebytes = p->width * 4;
  //tsize_t linebytes  = width * height * 4;

  
  snprintf(filename, 1024, "%s-%d/pixmap-%d-%lu.tiff", logdir, getpid(), id, p->p);

  // Don't overwrite
  if (stat(filename, &st) != -1) {
    return;
  }
  
  if ((img = XGetImage(display, p->p, 0, 0, p->width, p->height, AllPlanes, ZPixmap)) == NULL) {
    fprintf(stderr, "Failed XGetImage\n");
    return;
  }

  out = TIFFOpen(filename, "w");


  tmplVisualInfo.class=TrueColor;
  myVisualInfo=XGetVisualInfo(display,VisualClassMask,&tmplVisualInfo,&visuals);
  if (visuals==0)
    myVisualInfo=NULL;
  
  if (myVisualInfo!=NULL) {
      unsigned long maskFinder;
      int theShift;
      
      maskFinder=1; theShift=0;
      while ((maskFinder & myVisualInfo->red_mask)==0)
        {
          theShift++;
	  maskFinder=maskFinder << 1;
	};
      redShift=theShift-8+ones(myVisualInfo->red_mask);
      maskFinder=1; theShift=0;
      while ((maskFinder & myVisualInfo->green_mask)==0)
        {
          theShift++;
	  maskFinder=maskFinder << 1;
	};
      greenShift=theShift-8+ones(myVisualInfo->green_mask);
      maskFinder=1; theShift=0;
      while ((maskFinder & myVisualInfo->blue_mask)==0)
        {
          theShift++;
	  maskFinder=maskFinder << 1;
	};
      blueShift=theShift-8+ones(myVisualInfo->blue_mask);

  }
  
  TIFFSetField(out, TIFFTAG_IMAGEWIDTH, p->width);  // set the width of the image
  TIFFSetField(out, TIFFTAG_IMAGELENGTH, p->height);    // set the height of the image
  TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, sampleperpixel);   // set number of channels per pixel
  TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);    // set the size of the channels
  TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);    // set the origin of the image.
  //   Some other essential fields to set that you do not have to understand for now.
  TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
  
  
  
  unsigned char *buf = NULL;        // buffer used to store the row of pixel information for writing to file
  //    Allocating memory to store the pixels of current row
  if (TIFFScanlineSize(out) != linebytes) {
    linebytes = TIFFScanlineSize(out);
  }
  
  buf = (unsigned char *)_TIFFmalloc(linebytes);
  
  // We set the strip size of the file to be size of one row of pixels
  TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, p->width*sampleperpixel));
  
  
  
  //Now writing image to the file one strip at a time
  for (row = 0; row < p->height; row++) {
    int col;
    for (col=0; col< p->width; col++) {
      XColor swapColor;
      swapColor.pixel=XGetPixel(img,col,row);
      if (0) {
	//myVisualInfo==NULL) {
	//XQueryColor(MyDisplay,MColorMap,&swapColor);
	buf[col*3]=swapColor.red>>8;
	buf[col*3+1]=swapColor.green>>8;
	buf[col*3+2]=swapColor.blue>>8;
      } else {
	if (redShift>0) {
	  buf[col*3]=(swapColor.pixel & myVisualInfo->red_mask) >> redShift; 
	} else {
	  buf[col*3]=(swapColor.pixel & myVisualInfo->red_mask) << abs(redShift); 
	}
	
	if (greenShift>0) {
	  buf[col*3+1]=(swapColor.pixel & myVisualInfo->green_mask) >> greenShift; 
	} else {
	  buf[col*3+1]=(swapColor.pixel & myVisualInfo->green_mask) << abs(greenShift); 
	}
	
	if (blueShift>0) {
	  buf[col*3+2]=(swapColor.pixel & myVisualInfo->blue_mask) >> blueShift; 
	} else {
	  buf[col*3+2]=(swapColor.pixel & myVisualInfo->blue_mask) << abs(blueShift); 
	}
      }
    }
    
    //memcpy(buf, &img->data[row*linebytes], linebytes);
    // check the index here, and figure out why not using h*linebytes
    if (TIFFWriteScanline(out, buf, row, 0) < 0)
      break;
    
  }
  
  free(buf);
  TIFFClose(out);
  XDestroyImage(img);
}

void flushPixmaps(Display *d,
		  int id,
		  int force) {
  int j;
  for (j = 0 ; j < npixmaps[id] ; j++) {
    if (pixmaps[id][j].freed_at == 0 || force) {
      savePixmap(displays[id], id, &pixmaps[id][j]);
    }
  }  
}

void flushAllPixmaps(int force) {
  int i;
  for (i = 0 ; i < ndisplays ; i++) {
    flushPixmaps(displays[i], i, force);
  }
}

void doExit(void) {
  if (!save_at_checkpoint) {
    flushAllPixmaps(1);
  }
  if (logfile) {
    fclose(logfile);
    logfile = NULL;
  }
}

void doSignalUSR2(int signal) {
  exit(0);
}

void doSignalUSR1(int signal) {
  time_t now = time(NULL);
  fprintf(logfile, ". t=%ld\n", now);
  fflush(logfile);
  if (save_at_checkpoint) {
    flushAllPixmaps(0);
  }
}
